/* interp.c    interpreter of i-code   9 Jan 97
 *
 * port to PC on Turbo C 2.0		14 Oct 97
 * add timer interrupt 			20 Aug 98
 *
 * copyright(c) Prabhas Chongstitvatana
 */

#include "interp.h"
#include <dos.h>
/*
#include <conio.h>
*/

#define   execprim(c)   (*fntab[c])()

int  xLit(),
  xRvalg(),
  xLval(),
  xRval(),
  xFetch(),
  xJmp(),
  xJz(),
  xCall(),
  xSet(),
  xAdd(),
  xSub(),
  xMul(),
  xDiv(),
  xLt(),
  xLe(),
  xGe(),
  xGt(),
  xNe(),
  xEq(),
  xNot(),
  xAnd(),
  xOr(),
  xSend(),
  xReceive(),
  xRet0(),
  xRet1(),
  xPrint(),
  xPrintch(),
  xWait(),
  xSignal(),
  xFunc(),
  xProc(),
  xNop(),
  xStop(),
  xTmWait(),
  xTmSend(),
  xTmReceive(),
  xDelay(),
  xGettime()
;

#ifdef DEBUG
char  icname[][10] = {
  "Lit",
  "Lval",
  "Lvalg",
  "Rval",
  "Rvalg",
  "Fetch",
  "Set",
  "Index",
  "Jmp",
  "Jz",
  "Call",
  "Func",
  "Proc",
  "Ret0",
  "Ret1",
  "Stop",
  "Add",
  "Sub",
  "Mul",
  "Div",
  "Minus",
  "Not",
  "And",
  "Or",
  "Lt",
  "Le",
  "Eq",
  "Ne",
  "Ge",
  "Gt",
  "Print",
  "Printch",
  "Send",
  "Receive",
  "Wait",
  "Signal",
  "Nop",
  "tmSend",
  "tmReceive",
  "tmWait",
  "Delay",
  "Gettime"
} ;
#endif

fnptr fntab[] = {
  xLit,
  xLval,
  /* xLvalg, */ 	xLit,
  xRval,
  xRvalg,
  xFetch,
  xSet,
  /* xIndex */ 		xAdd,
  xJmp,
  xJz,
  xCall,
  xFunc,
  xProc,
  xRet0,
  xRet1,
  xStop,
  xAdd,
  xSub,
  xMul,
  xDiv,
  /* xMinus */  	xNop,
  xNot,
  xAnd,
  xOr,
  xLt,
  xLe,
  xEq,
  xNe,
  xGe,
  xGt,
  xPrint,
  xPrintch,
  xSend,
  xReceive,
  xWait,
  xSignal,
  xNop,
  xTmSend,
  xTmReceive,
  xTmWait,
  xDelay,
  xGettime
};

int maxsp,sp,fp,ip;        /* current state of computation */
int M[MAXMEM];             /* data and stack segment memory */
uchar cs[MAXCS];           /* code segment */
FILE *FI;		   /* input icode file */

extern int flagtimeout;
extern tmp tmlist;

/* i-code interpreter */
void interp()
{
  do {
    while(activep != nil) {	/* active process loop */
      execprim(cs[ip]);

#ifdef DEBUG
      trace();
      if( pclock++ > timelimit ) {
	pclock = 0;
	switchp();
      }
#else
      if( flagtimeout ) {
	flagtimeout = 0;
	switchp();
      }
#endif
    } 	/* end while */
	/* outer loop wait for timer */
#ifndef DEBUG
    delay(20);
#endif
    vclock++;
    dotimer();
  } while ( tmlist != nil ) ;
}

#ifdef DEBUG
trace()
{
  int i;
  printf("ip %d ",ip);
  if( cs[ip] > VALIDOP ) printf("op Undef ");
  else printf("op %s TOS ",icname[cs[ip]]);
  for( i=sp; i>sp-8; i-- ) {
    if( i <= DSSIZE ) printf(". ");
    else if( i == fp+2 ) printf("[%d ",M[i]);
    else if( i == fp ) printf("%d] ",M[i]);
    else printf("%d ",M[i]);
  }
  printf("\n");
}
#endif

error(mess)
char *mess;
{
  printf("%s at ip %d\n",mess,ip);
  exit(0);
}

op1() 	{ return ( cs[ip+1]*256 + cs[ip+2] ); }
op2() 	{ return ( cs[ip+3]*256 + cs[ip+4] ); }
op3() 	{ return ( cs[ip+5]*256 + cs[ip+6] ); }

push(x) int x; { sp++; M[sp] = x; }
pop() 	{ return( M[sp--] ); }

xLit() 	{ push( op1()); ip += 3; }
xRvalg(){ push( M[op1()]); ip += 3; }
xLval() { push( fp-op1()); ip += 3; }
xRval() { push( M[fp-op1()]); ip += 3; }
xFetch(){ push( M[pop()]); ip++; }
xJmp()	{ ip = op1(); }
xJz()	{ if( pop() == FALSE) ip = op1(); else ip += 3; }
xCall() { push( ip+3 ); ip = op1(); }
xSet()  { M[NEXT] = TOS; pop2; ip++; }

xAdd()	{ int a; a = NEXT + TOS; pop2; push(a); ip++; }
xSub()	{ int a; a = NEXT - TOS; pop2; push(a); ip++; }
xMul() 	{ int a; a = NEXT * TOS; pop2; push(a); ip++; }
xDiv() 	{ int a; a = NEXT / TOS; pop2; push(a); ip++; }

xLt()	{ int a; a = NEXT <  TOS; pop2; push(a); ip++; }
xLe()	{ int a; a = NEXT <= TOS; pop2; push(a); ip++; }
xGt()	{ int a; a = NEXT >  TOS; pop2; push(a); ip++; }
xGe()	{ int a; a = NEXT >= TOS; pop2; push(a); ip++; }
xNe()	{ int a; a = NEXT != TOS; pop2; push(a); ip++; }
xEq()	{ int a; a = NEXT == TOS; pop2; push(a); ip++; }
xAnd()	{ int a; a = NEXT && TOS; pop2; push(a); ip++; }
xOr()	{ int a; a = NEXT || TOS; pop2; push(a); ip++; }
xNot()  { push( ! pop() ); ip++; }

xRet0()	{ ip = M[fp+2]; sp = M[fp+1]; fp = M[fp]; }
xRet1()
{
  int a;
  a = TOS;
  ip = M[fp+2]; sp = M[fp+1]; fp = M[fp];
  push(a);
}

xPrint()   { printf("%d",pop()); ip++; }
xPrintch() { printf("%c",pop()); ip++; }
xFunc()
{
  int fp1,a,b;
  a = op1();
  b = op2();
  if( sp+b+FUDGE > maxsp ) error("stack overflow");
  fp1 = sp + b;
  M[fp1+2] = TOS;            /* assume caller push ret ads */
  M[fp1+1] = sp - a - 1;
  M[fp1] = fp;
  fp = fp1;
  sp = fp1 + 2;
  ip += 5;
}

xNop() 	{ error("opcode not implemented"); }
xStop() { blockp(TERMINATED); }

int icodesize;

/* Note:  On DOS, using getc from stdin will lost the char 13
   therefore have to get from file here  (Turbo C 2.0)
   On Unix, this is not happening.       (cc )
 */
loadIcode()
{
  int c;
  cs[0] = 0;
  icodesize = 1;
  while( (c = fgetc(FI)) != EOF)
    cs[icodesize++] = c;
}

chkIcode()
{
  int csize, dsize, psize, f;
  readHead(FI,&csize,&dsize,&psize);
  f = 0;
  if( csize > MAXCS ) {
    printf("code size too large : %d\n",csize);
    f = 1;
  }
  if( dsize > DSSIZE ) {
    printf("data size too large : %d\n",dsize);
    f = 1;
  }
  if( psize > maxprocess ) {
    printf("too many process : %d\n",psize);
    f = 1;
  }
  if( f ) {
    printf("check interpreter limit in interp.h\n");
    exit(-1);
  }
}

coredump()
{
  int i;
  for(i=1; i<icodesize; i++) printf("%d ",cs[i]);
}

main(argc, argv)
int argc;
char * argv[];
{
  if( argc < 2 ) {
    printf("usage : interp icode");
    exit(-1);
  }
  FI = fopen(argv[1],"rb");
  if( FI == NULL ) {
    printf("cannot open input : %s\n",argv[1]);
    exit(-1);
  }
  chkIcode();
  loadIcode();
  fclose(FI);
  initmail();
  inittimer();
  startup();
  interp();
#ifndef DEBUG
  restore_isr();
#endif
  printf("\nstop\n");
}

